from typing import Optional

import torch


def banded_square_matrix(
    ndim: int,
    bandwidth: int,
):
    """
    Generate a banded square matrix with entries 0 or 1.
    The rows wrap around, so the number of non-zero entries in each row is 2 * bandwidth + 1.

    Args:
    ndim: An integer representing the dimension of the square matrix.
    bandwidth: An integer representing the bandwidth of the matrix.

    Returns:
    A tensor of shape (ndim, ndim) representing the banded square matrix.
    """
    matrix = torch.zeros(ndim, ndim)
    for i in range(ndim):
        for j in range(i - bandwidth, i + bandwidth + 1):
            matrix[i, j % ndim] = 1
    return matrix


def dmp_marginal_dist(
    initial_dist: torch.Tensor,
    transition_matrix: torch.Tensor,
    num_steps: int,
):
    """Compute the marginal distribution of a discrete Markov process.

    Args:
    initial_dist: A tensor of shape (n, 1) representing the initial distribution.
    transition_matrix: A tensor of shape (n, n) representing the transition matrix.
    num_steps: An integer representing the number of steps to take.

    Returns:
    A tensor of shape (n, 1) representing the marginal distribution after num_steps steps.
    """
    marginal_dist = initial_dist
    for _ in range(num_steps):
        marginal_dist = torch.matmul(transition_matrix, marginal_dist)
    return marginal_dist


def dmp_stationary_dist(
    transition_matrix: torch.Tensor,
    tolerance: float = 1e-5,
):
    """Compute the stationary distribution of a discrete Markov process.

    Args:
    transition_matrix: A tensor of shape (n, n) representing the transition matrix.

    Returns:
    A tensor of shape (n, 1) representing the stationary distribution.
    """
    n = transition_matrix.shape[0]
    A = (torch.eye(n) - transition_matrix).T
    A = torch.cat([A, torch.ones(n).reshape(1, -1)], dim=0)

    b = torch.zeros(n + 1, 1)
    b[-1] = 1

    try:
        # get QR decomposition
        Q, R = torch.linalg.qr(A)
        soln = torch.linalg.solve(R, Q.T @ b)
        resid = torch.norm(A @ soln - b) / torch.norm(b)
        if resid > tolerance:
            raise torch._C._LinAlgError
    except torch._C._LinAlgError as err:
        msg = "Stationary distribution solve failed. " "Are you sure the stationary distribution exists?"
        raise RuntimeError(msg) from err

    return soln


def sample_sparse_ergodic_transition_matrix(
    num_states: int,
    bandwidth: int,
    softmax_temp: float = 1.0,
    generator: Optional[torch.Generator] = None,
    repeats_always_possible: bool = False,
):
    """
    Sample a transition matrix $P$ that satisfies the ergodicity conditions
    - irreducibility
    - aperiodicity
    - positive recurrence
    Additionally, some entries of the transition matrix must be zero.
    """

    if softmax_temp <= 0:
        msg = "Softmax temperature must be greater than 0."
        raise ValueError(msg)

    # set generator
    if generator is None:
        generator = torch.Generator()

    randn_matrix = torch.randn(num_states, num_states, generator=generator)
    print(randn_matrix)
    dense_transition_matrix = (randn_matrix / softmax_temp).softmax(dim=-1)
    print(dense_transition_matrix)

    # construct mask as banded matrix
    mask = banded_square_matrix(num_states, bandwidth).bool()
    # shuffle the mask
    mask = mask[torch.randperm(num_states, generator=generator)]

    if repeats_always_possible:
        # set diagonal entries of mask to True
        mask = mask | torch.eye(num_states, dtype=torch.bool)

    print(mask)

    transition_matrix = torch.where(mask, dense_transition_matrix, torch.zeros_like(dense_transition_matrix))
    transition_matrix = transition_matrix / transition_matrix.sum(dim=-1, keepdim=True)

    # check Perron-Frobenius theorem
    m = (num_states - 1) ** 2 + 1

    # compute P^m
    transition_matrix_m = transition_matrix
    for _ in range(m - 1):
        transition_matrix_m = transition_matrix_m @ transition_matrix

    if not torch.all(transition_matrix_m > 0):
        msg = "Perron-Frobenius theorem not satisfied."
        raise RuntimeError(msg)

    return transition_matrix


def sample_dmp(
    initial_dist: torch.Tensor,
    transition_matrix: torch.Tensor,
    num_steps: int,
    num_samples: int = 1,
    generator: Optional[torch.Generator] = None,
):
    samples = torch.zeros(num_samples, num_steps, dtype=torch.int64)
    if generator is None:
        generator = torch.Generator(device=initial_dist.device)
    samples[:, 0] = torch.multinomial(
        initial_dist,
        num_samples,
        replacement=True,
        generator=generator,
    )
    for t in range(1, num_steps):
        # import pdb; pdb.set_trace()
        samples[:, t] = torch.multinomial(
            transition_matrix[samples[:, t - 1]],
            1,
            generator=generator,
        ).squeeze(-1)
    return samples.to(initial_dist.device)


def dmp_sample_log_likelihood(
    samples: torch.Tensor,
    initial_dist: torch.Tensor,
    transition_matrix: torch.Tensor,
):
    samples = samples.long()  # may need to cast from float
    log_likelihood = 0.0
    log_likelihood += torch.log(initial_dist[samples[:, 0]])
    for t in range(1, samples.size(1)):
        log_likelihood += torch.log(transition_matrix[samples[:, t - 1], samples[:, t]])
    return log_likelihood



import pytest
import torch

from holo.test_functions.closed_form import (
    Ehrlich,
)
from holo.test_functions.elemental import (
    dmp_stationary_dist,
    sample_dmp,
)


def test_ehrlich_single_motif():
    num_states = 32
    num_steps = 256
    motif_length = 8
    noise_std = 0.0
    negate = False
    random_seed = 0
    ehrlich = Ehrlich(
        num_states=num_states,
        dim=num_steps,
        motif_length=motif_length,
        noise_std=noise_std,
        negate=negate,
        random_seed=random_seed,
    )
    num_samples = 4
    stationary_dist = dmp_stationary_dist(ehrlich.transition_matrix)
    generator = torch.Generator().manual_seed(random_seed)
    dmp_samples = sample_dmp(
        initial_dist=stationary_dist.squeeze(),
        transition_matrix=ehrlich.transition_matrix,
        num_steps=ehrlich.dim,
        num_samples=num_samples,
        generator=generator,
    )
    f = ehrlich.evaluate_true(dmp_samples)
    assert torch.all(f >= 0.0)
    assert torch.all(f <= 1.0)

    unif_samples = torch.randint(0, num_states, (num_samples, num_steps))
    f = ehrlich.evaluate_true(unif_samples)
    assert torch.allclose(f, torch.full_like(f, -float("inf")))


def test_ehrlich_multi_motif():
    num_states = 32
    num_steps = 256
    motif_length = 8
    random_seed = 0
    ehrlich = Ehrlich(
        num_states=num_states,
        dim=num_steps,
        num_motifs=4,
        motif_length=motif_length,
        random_seed=random_seed,
    )
    num_samples = 4
    stationary_dist = dmp_stationary_dist(ehrlich.transition_matrix)
    generator = torch.Generator().manual_seed(random_seed)
    dmp_samples = sample_dmp(
        initial_dist=stationary_dist.squeeze(),
        transition_matrix=ehrlich.transition_matrix,
        num_steps=ehrlich.dim,
        num_samples=num_samples,
        generator=generator,
    )
    f = ehrlich.evaluate_true(dmp_samples)
    assert torch.all(f >= 0.0)
    assert torch.all(f <= 1.0)

    unif_samples = torch.randint(0, num_states, (num_samples, num_steps))
    f = ehrlich.evaluate_true(unif_samples)
    assert torch.allclose(f, torch.full_like(f, -float("inf")))


def test_invalid_ehrlich():
    with pytest.raises(ValueError):
        Ehrlich(
            num_states=32,
            dim=2,
            num_motifs=1,
            motif_length=8,
        )

    with pytest.raises(ValueError):
        Ehrlich(
            num_states=32,
            dim=16,
            num_motifs=4,
            motif_length=8,
            random_seed=0,
        )



import math
from typing import Optional

import torch


def motif_search(
    solution: torch.LongTensor,
    motif: torch.LongTensor,
    spacing: Optional[torch.LongTensor] = None,
    mode: str = "present",
    quantization: Optional[int] = None,
):
    """
    Check if a spaced motif is present in a solution.
    The elements of the motif should be spaced according to `spacing`.
    If spacing is not provided, it is assumed to be 1.
    If `mode` is "count", the number of motifs in the solution is returned.
    If `strict` is True, only motifs that are fully satisfied are counted, otherwise
        the fraction of the motif that is satisfied is returned.
    """
    motif_size = motif.size(-1)

    if quantization is None:
        quantization = motif_size

    if spacing is None:
        size = solution.size()[:-1]
        spacing = torch.ones(
            *size,
            motif_size - 1,
            dtype=torch.long,
            device=solution.device,
        )
    else:
        size = solution.size()[:-1]
        spacing = spacing.expand(*size, -1)

    # convert spacing into index tensor for gather
    base_index = torch.cat([torch.zeros_like(spacing[..., 0]).unsqueeze(-1), spacing.cumsum(-1)], dim=-1)

    # slide indices out to maximum length
    max_base = base_index.max()
    num_steps = solution.size(-1) - max_base
    index_delta = torch.arange(num_steps, device=solution.device)
    expand_args = [1] * len(base_index.shape)
    index_delta = index_delta.view(-1, *expand_args)
    index = base_index + index_delta

    # gather solution subsequences
    gathered = torch.stack([torch.gather(solution, -1, step) for step in index])

    # check if motif is present
    is_equal = gathered.eq(motif)
    present_count = is_equal.float().sum(-1)

    # if strict, only count motifs that are fully satisfied
    # if strict:
    #     frac_satisfied = frac_satisfied.eq(1.0).float()
    quant_factor_1 = math.ceil(motif_size / quantization)
    quant_factor_2 = motif_size / quant_factor_1
    quantized_count = present_count // quant_factor_1 / quant_factor_2

    if mode == "count":
        return quantized_count.sum(0)

    return quantized_count.max(0).values



import torch
from botorch.test_functions import SyntheticTestFunction

from holo.test_functions.elemental import hamming_dist


class RoughMtFuji(SyntheticTestFunction):
    """
    Implementation of the Rough Mt. Fuji fitness landscape
    as described in Neidhart et al. (2014)
    'Adaptation in tunably rugged fitness landscapes:
    The Rough Mount Fuji model'
    https://arxiv.org/abs/1402.3065
    """

    num_states = 2
    num_centroids = 1

    def __init__(
        self,
        dim: int = 2,
        additive_term: float = 0.25,
        random_term_std: float = 1.0,
        noise_std: float = 0.0,
        negate: bool = False,
        random_seed: int = 0,
    ):
        self.dim = dim
        super(RoughMtFuji, self).__init__(
            noise_std=noise_std,
            negate=negate,
            bounds=[(0.0, 1.0) for _ in range(dim)],
        )
        self._random_seed = random_seed
        self._generator = torch.Generator().manual_seed(random_seed)
        self.centroids = torch.randint(0, 2, (1, dim), generator=self._generator)
        self._additive_term = additive_term
        self._random_term_std = random_term_std
        self._random_term = torch.randn(1, dim, generator=self._generator) * random_term_std

    def evaluate_true(self, X: torch.Tensor) -> torch.Tensor:
        dist = hamming_dist(X, self.centroids, dim=-1)
        return -self._additive_term * dist + (self._random_term * X).sum(dim=-1)

    def initial_solution(self, n: int = 1):
        # reset generator seed so initial solution is always the same
        self._generator = self._generator.manual_seed(self._random_seed)
        return self.random_solution(n)

    def random_solution(self, n: int = 1):
        unif_samples = torch.randint(self.num_states, (n, self.dim), device=self.centroids.device)
        if n == 1:
            return unif_samples.squeeze(0)
        return unif_samples

    @property
    def _optimal_value(self):
        return self.evaluate_true(self.optimal_solution()).squeeze()

    def optimal_solution(self):
        soln = self.centroids.clone()
        mask = self._random_term - self._additive_term > 0
        soln = torch.where(mask, torch.ones_like(soln), soln)
        mask = self._random_term + self._additive_term < 0
        soln = torch.where(mask, torch.zeros_like(soln), soln)

        return soln

    def to(self, device, dtype):
        self.centroids = self.centroids.to(device, dtype)
        self._generator = torch.Generator(device=device).manual_seed(self._random_seed)
        return self

    def __repr__(self):
        return (
            f"RoughMtFuji("
            f"num_states={self.num_states}, "
            f"dim={self.dim}, "
            f"additive_term={self._additive_term}, "
            f"random_term={self._random_term}, "
            f"random_term_std={self._random_term_std}, "
            f"centroids={self.centroids}, "
            f"noise_std={self.noise_std}, "
            f"negate={self.negate}, "
            f"random_seed={self._random_seed})"
        )



import torch
from botorch.test_functions import SyntheticTestFunction

from holo.test_functions.elemental import (
    dmp_sample_log_likelihood,
    dmp_stationary_dist,
    motif_search,
    sample_dmp,
    sample_sparse_ergodic_transition_matrix,
)


class Ehrlich(SyntheticTestFunction):
    _optimal_value = 1.0
    num_objectives = 1

    def __init__(
        self,
        num_states: int = 5,
        dim: int = 7,
        num_motifs: int = 1,
        motif_length: int = 3,
        quantization: int | None = None,
        noise_std: float = 0.0,
        negate: bool = False,
        random_seed: int = 0,
    ) -> None:
        bounds = [(0.0, float(num_states - 1)) for _ in range(dim)]
        self.num_states = num_states
        self.dim = dim
        self._random_seed = random_seed
        self._quantization = quantization
        super(Ehrlich, self).__init__(
            noise_std=noise_std,
            negate=negate,
            bounds=bounds,
        )
        self._generator = torch.Generator().manual_seed(random_seed)
        self.initial_dist = torch.ones(num_states) / num_states
        bandwidth = int(num_states * 0.4)
        self.transition_matrix = sample_sparse_ergodic_transition_matrix(
            num_states, bandwidth, softmax_temp=0.5, generator=self._generator, repeats_always_possible=True
        )
        self.stationary_dist = dmp_stationary_dist(self.transition_matrix)

        slack_positions = dim - num_motifs * motif_length
        element_gaps = num_motifs * (motif_length - 1)
        max_spacing = 1 + slack_positions // element_gaps
        if max_spacing < 1:
            raise ValueError("cannot guarantee a solution satisfying all motifs exists.")

        # draw motifs as single sequence from DMP and chunk to ensure feasible soln exists
        all_motifs = sample_dmp(
            initial_dist=self.stationary_dist.squeeze(-1),
            transition_matrix=self.transition_matrix,
            num_steps=num_motifs * motif_length,
            num_samples=1,
            generator=self._generator,
        ).squeeze(0)
        self.motifs = torch.chunk(all_motifs, num_motifs, dim=0)

        self.spacings = []
        for _ in range(num_motifs):
            # draw random spacing
            # spacing = torch.randint(1, max_spacing + 1, (motif_length - 1,), generator=self._generator)
            # random draw from (motif_length - 1) simplex
            weights = torch.rand(motif_length - 1, generator=self._generator)
            weights /= weights.sum()
            spacing = (slack_positions // num_motifs) * weights
            # round down
            spacing = 1 + spacing.floor().to(torch.int64)
            self.spacings.append(spacing)

    def evaluate_true(self, X: torch.Tensor) -> torch.Tensor:
        motif_present = []
        for motif, spacing in zip(self.motifs, self.spacings):
            motif_present.append(
                motif_search(
                    solution=X,
                    motif=motif,
                    spacing=spacing,
                    mode="present",
                    quantization=self._quantization,
                )
            )
        all_motifs_present = torch.stack(motif_present).prod(dim=0)
        log_likelihood = dmp_sample_log_likelihood(
            samples=X,
            initial_dist=self.initial_dist,
            transition_matrix=self.transition_matrix,
        )
        is_feasible = log_likelihood > -float("inf")
        return torch.where(is_feasible, all_motifs_present, -float("inf"))

    def initial_solution(self, n: int = 1):
        # reset generator seed so initial solution is always the same
        self._generator = self._generator.manual_seed(self._random_seed)
        dmp_samples = sample_dmp(
            initial_dist=self.stationary_dist.squeeze(-1),
            transition_matrix=self.transition_matrix,
            num_steps=self.dim,
            num_samples=n,
            generator=self._generator,
        )
        if n == 1:
            return dmp_samples.squeeze(0)
        return dmp_samples

    def random_solution(self, n: int = 1):
        unif_samples = torch.randint(self.num_states, (n, self.dim), device=self.initial_dist.device)
        if n == 1:
            return unif_samples.squeeze(0)
        return unif_samples

    def optimal_solution(self):
        # sample random sequence from DMP
        soln = torch.zeros(self.dim, dtype=torch.int64, device=self.initial_dist.device)
        # fill in spaced motifs with repeats
        position = 0
        for motif, spacing in zip(self.motifs, self.spacings):
            spacing = torch.cat(
                [
                    torch.tensor([0], device=self.initial_dist.device),
                    spacing,
                ]
            )
            index = spacing.cumsum(0).tolist()
            print(index)
            motif = motif.tolist()
            for idx in range(index[-1] + 1):
                if idx in index:
                    next_state = motif.pop(0)
                soln[position] = next_state
                # print(position)
                position += 1
        # fill in remaining states with last state of last motif
        soln[position:] = soln[position - 1]

        # check optimal value
        optimal_value = self.evaluate_true(soln.unsqueeze(0))
        if not optimal_value == self._optimal_value:
            print(soln)
            print(optimal_value)
            raise RuntimeError("optimal value not achieved by optimal solution.")
        return soln

    def to(self, device, dtype):
        self.transition_matrix = self.transition_matrix.to(device, dtype)
        self.initial_dist = self.initial_dist.to(device, dtype)
        self.stationary_dist = self.stationary_dist.to(device, dtype)
        self.motifs = [motif.to(device) for motif in self.motifs]
        self.spacings = [spacing.to(device) for spacing in self.spacings]
        self._generator = torch.Generator(device=device).manual_seed(self._random_seed)
        return self

    def __repr__(self):
        motif_list = [f"motif_{i}: {motif.tolist()}" for i, motif in enumerate(self.motifs)]
        spacing_list = [f"spacing_{i}: {spacing.tolist()}" for i, spacing in enumerate(self.spacings)]
        return (
            f"Ehrlich("
            f"num_states={self.num_states}, "
            f"dim={self.dim}, "
            f"num_motifs={len(self.motifs)}, "
            f"motifs=[{', '.join(motif_list)}], "
            f"spacings=[{', '.join(spacing_list)}], "
            f"quantization={self._quantization}, "
            f"noise_std={self.noise_std}, "
            f"negate={self.negate}, "
            f"random_seed={self._random_seed})"
        )



import torch

from holo.optim import DiscreteEvolution
from holo.test_functions.closed_form import Ehrlich


def test_discrete_evolution_single_param():
    seq_len = 3
    vocab_size = 5
    num_steps = 20
    mutation_prob = 0.2
    recombine_prob = 0.5
    params = [
        torch.nn.Parameter(
            torch.tensor([vocab_size - 1] * seq_len).float(),
        )
    ]
    vocab = list(range(vocab_size))
    optimizer = DiscreteEvolution(params, vocab, mutation_prob=mutation_prob, recombine_prob=recombine_prob)

    def closure(param_list):
        return param_list[0].sum(-1)

    for _ in range(num_steps):
        loss = optimizer.step(closure)

    assert torch.allclose(loss, torch.zeros_like(loss))
    assert torch.allclose(params[0].data, torch.zeros_like(params[0].data))


def test_discrete_evolution_param_groups():
    seq_lens = [1, 3]
    vocab_size = 5
    num_steps = 20

    param_groups = [
        {
            "params": [
                torch.nn.Parameter(
                    torch.tensor([vocab_size - 1] * seq_lens[0]).float(),
                )
            ],
            "mutation_prob": 0.9,
            "recombine_prob": 0.1,
        },
        {
            "params": [
                torch.nn.Parameter(
                    torch.tensor([vocab_size - 1] * seq_lens[1]).float(),
                )
            ],
            "mutation_prob": 0.1,
            "recombine_prob": 0.9,
        },
    ]

    vocab = list(range(vocab_size))
    optimizer = DiscreteEvolution(param_groups, vocab)

    def closure(param_list):
        return sum(p.sum(-1) for p in param_list)

    for _ in range(num_steps):
        loss = optimizer.step(closure)

    assert torch.allclose(loss, torch.zeros_like(loss))
    param_list = [p.data for group in param_groups for p in group["params"]]
    cat_param = torch.cat(param_list, dim=-1)
    assert torch.allclose(cat_param, torch.zeros_like(cat_param))


def test_ehrlich_optimization():
    num_states = 32
    num_steps = 128
    motif_length = 4
    noise_std = 0.0
    random_seed = 0
    ehrlich = Ehrlich(
        num_states=num_states,
        dim=num_steps,
        num_motifs=1,
        motif_length=motif_length,
        noise_std=noise_std,
        negate=True,
        random_seed=random_seed,
    )

    # initialization
    initial_solution = ehrlich.initial_solution()

    def closure(param_list):
        return ehrlich(param_list[0])

    params = [torch.nn.Parameter(initial_solution.float())]

    mutation_prob = 1 / num_steps
    recombine_prob = 1 / num_steps
    survival_quantile = 0.01
    num_particles = 2048
    optimizer = DiscreteEvolution(
        params,
        vocab=list(range(num_states)),
        mutation_prob=mutation_prob,
        recombine_prob=recombine_prob,
        survival_quantile=survival_quantile,
        num_particles=num_particles,
    )
    num_opt_steps = 256

    best = float("inf")
    for t in range(num_opt_steps):
        loss = optimizer.step(closure)
        if loss < best:
            best = loss.item()

    assert best == -1.0



from typing import MutableMapping

import wandb
from omegaconf import DictConfig, OmegaConf

import holo


def wandb_setup(cfg: DictConfig):
    """
    Runs `wandb.init` and `wandb.login`.
    The values in `cfg` are logged to the wandb run.
    """
    if not hasattr(cfg, "wandb_host"):
        cfg["wandb_host"] = "https://api.wandb.ai"

    if not hasattr(cfg, "wandb_mode"):
        cfg["wandb_mode"] = "online"

    if not hasattr(cfg, "project_name"):
        cfg["project_name"] = "holo"

    if not hasattr(cfg, "exp_name"):
        cfg["exp_name"] = "default_group"

    wandb.login(host=cfg.wandb_host)

    wandb.init(
        project=cfg.project_name,
        mode=cfg.wandb_mode,
        group=cfg.exp_name,
    )
    cfg["job_name"] = wandb.run.name
    cfg["__version__"] = holo.__version__
    log_cfg = flatten_config(OmegaConf.to_container(cfg, resolve=True))
    wandb.config.update(log_cfg)


def flatten_config(d: DictConfig, parent_key: str = "", sep: str = "/"):
    """
    Flatten a nested `DictConfig` object into a flat dictionary for logging.
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten_config(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)



from typing import Iterable

import torch
from torch.optim import Optimizer


class DiscreteEvolution(Optimizer):
    r"""Implements discrete evolution optimizer."""

    def __init__(
        self,
        params: Iterable[torch.nn.Parameter],
        vocab: Iterable[list[int]],
        mutation_prob: float = 0.2,
        recombine_prob: float = 0.5,
        num_particles: int = 128,
        survival_quantile: float = 0.5,
    ):
        """
        Args:
            params: Iterable of parameters to optimize or dicts defining parameter groups.
            vocab: Iterable of lists of integers representing the vocabulary for each parameter group.
            mutation_prob: Probability of mutating each element of the parameters.
            recombine_prob: Probability of recombining elements of the parameters.
            num_particles: Number of particles to maintain for each parameter.
            survival_quantile: Fraction of particles to keep at each step.
        """
        if mutation_prob <= 0.0 or mutation_prob >= 1.0:
            msg = "Mutation probability must be in (0, 1)."
            raise ValueError(msg)

        if recombine_prob < 0.0 or recombine_prob >= 1.0:
            msg = "Recombination probability must be in [0, 1)."
            raise ValueError(msg)

        if num_particles < 1:
            msg = "Number of particles must be positive."
            raise ValueError(msg)

        if survival_quantile <= 0.0 or survival_quantile >= 1.0:
            msg = "Survival quantile must be in (0, 1)."
            raise ValueError(msg)

        if survival_quantile * num_particles < 2:
            msg = "Survival quantile must be large enough to keep at least two particles."
            raise ValueError(msg)

        defaults = dict(
            mutation_prob=mutation_prob,
            recombine_prob=recombine_prob,
            vocab=vocab,
        )

        super(DiscreteEvolution, self).__init__(params, defaults)
        self._num_particles = num_particles
        self._survival_quantile = survival_quantile
        self.particle_loss = None

    def _init_group(self, group, particle_buffer_list):
        for p in group["params"]:
            state = self.state[p]
            if "particles" not in state:
                repeat_args = [1] * len(p.data.shape)
                particles = p.data.clone().repeat(self._num_particles, *repeat_args)
                particles.requires_grad_(False)
                state["particles"] = particles
                particle_buffer_list.append(particles)
            else:
                particle_buffer_list.append(state["particles"])

    def step(self, closure: callable):
        particle_buffer_list = []
        for group in self.param_groups:
            self._init_group(group, particle_buffer_list)

        self.particle_loss = closure(particle_buffer_list)  # (num_particles,)
        particle_arg_min = torch.argmin(self.particle_loss)
        soln_loss = self.particle_loss[particle_arg_min]

        # update parameters to current best
        for group in self.param_groups:
            for p in group["params"]:
                state = self.state[p]
                particles = state["particles"]
                p.data = particles[particle_arg_min].clone()

        # update particles
        # particle survival threshold determined by quantile
        threshold = torch.quantile(self.particle_loss, self._survival_quantile)
        if threshold < float("inf"):
            surviving = self.particle_loss <= threshold
        else:
            surviving = torch.ones_like(self.particle_loss, dtype=torch.bool)
        for group in self.param_groups:
            num_params = len(group["params"])
            count = 0
            while count < num_params:
                particles = particle_buffer_list.pop(0)
                count += 1

                surviving_particles = particles[surviving]
                num_surviving = surviving_particles.size(0)
                num_missing = self._num_particles - num_surviving

                # repopulate missing particles by recombining survivors
                parent_1 = torch.randint(0, num_surviving, (num_missing,))
                parent_2 = torch.randint(0, num_surviving, (num_missing,))
                recombine_prob = group["recombine_prob"]
                recombine_mask = torch.rand_like(surviving_particles[parent_1]) < recombine_prob
                replace_particles = torch.where(
                    recombine_mask,
                    surviving_particles[parent_1],
                    surviving_particles[parent_2],
                )
                particles[~surviving] = replace_particles

                # mutate particles
                mutation_mask = torch.rand_like(particles) < group["mutation_prob"]
                vocab_tensor = torch.tensor(group["vocab"], device=particles.device)
                replace_elements = torch.randint(
                    len(vocab_tensor),
                    particles.shape,
                    device=particles.device,
                    dtype=particles.dtype,
                )
                new_particles = torch.where(mutation_mask, replace_elements, particles)
                particles.copy_(new_particles)

        return soln_loss



import torch

from holo.test_functions.elemental import motif_search


def test_motif_search():
    cases = torch.tensor(
        [
            [0, 1, 2, 3, 4],
            [0, 1, 2, 4, 5],
            [0, 1, 4, 5, 6],
            [0, 4, 5, 6, 7],
            [4, 5, 6, 7, 8],
        ]
    )
    motif = cases[0, :-1]

    res = motif_search(
        solution=cases,
        motif=motif,
        spacing=None,
        mode="present",
        quantization=1,
    )
    assert torch.all(res == torch.tensor([1.0, 0.0, 0.0, 0.0, 0.0]))

    res = motif_search(
        solution=cases,
        motif=motif,
        spacing=None,
        mode="present",
        quantization=4,
    )
    assert torch.all(res == torch.tensor([1.0, 0.75, 0.5, 0.25, 0.0]))


def test_spaced_motif_search():
    cases = torch.tensor(
        [
            [0, 1, 2, 3, 4],
            [0, 1, 2, 4, 5],
            [0, 1, 4, 5, 6],
            [0, 4, 5, 6, 7],
            [4, 5, 6, 7, 8],
        ]
    )
    motif = cases[0, ::2]
    spacing = torch.tensor([2, 2])
    res = motif_search(
        solution=cases,
        motif=motif,
        spacing=spacing,
        mode="present",
        quantization=1,
    )
    assert torch.all(res == torch.tensor([1.0, 0.0, 0.0, 0.0, 0.0]))



from ._discrete_markov_process import (
    banded_square_matrix,
    dmp_marginal_dist,
    dmp_sample_log_likelihood,
    dmp_stationary_dist,
    sample_dmp,
    sample_sparse_ergodic_transition_matrix,
)
from ._hamming_dist import hamming_dist
from ._motif import motif_search



import random

import hydra
import numpy as np
import torch
from omegaconf import OmegaConf

import wandb
from holo.logging import wandb_setup


@hydra.main(version_base=None, config_path="../config/hydra", config_name="benchmark_optimizer")
def main(cfg):
    if cfg.optimizer.mutation_prob is None:
        cfg.optimizer.mutation_prob = 1.1 / cfg.test_function.dim
    if cfg.optimizer.recombine_prob is None:
        cfg.optimizer.recombine_prob = 1.1 / cfg.test_function.dim

    wandb_setup(cfg)
    random.seed(cfg.random_seed)
    np.random.seed(cfg.random_seed)
    torch.manual_seed(cfg.random_seed)
    print(OmegaConf.to_yaml(cfg, resolve=True))

    dtype = torch.double if cfg.dtype == "float64" else torch.float
    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

    test_function = hydra.utils.instantiate(cfg.test_function)
    test_function = test_function.to(device, dtype)

    initial_solution = test_function.initial_solution().to(dtype)
    params = [torch.nn.Parameter(initial_solution)]

    print(f"Test function: {test_function}")
    known_soln = test_function.optimal_solution()
    opt_val = test_function.optimal_value
    print(f"Known optimal solution: {known_soln}")

    def closure(param_list):
        return test_function(param_list[0])

    vocab = list(range(test_function.num_states))
    optimizer = hydra.utils.instantiate(cfg.optimizer, params=params, vocab=vocab)

    print(f"Searching for solution with optimal value {opt_val}...")
    cumulative_regret = 0.0
    best_loss = float("inf")
    for t_idx in range(cfg.num_opt_steps):
        loss = optimizer.step(closure)
        if loss < best_loss:
            best_loss = loss.item()
            best_params = [p.data.clone() for p in params]

        simple_regret_best = best_loss - opt_val
        simple_regret_last = loss - opt_val
        cumulative_regret += best_loss - opt_val
        # frac_particles_feasible = optimizer.particle_loss.gt(-float("inf")).float().mean().item()
        frac_particles_feasible = optimizer.particle_loss.lt(float("inf")).float().mean().item()

        metrics = {
            "simple_regret_best": simple_regret_best,
            "simple_regret_last": simple_regret_last,
            "cumulative_regret": cumulative_regret,
            "frac_particles_feasible": frac_particles_feasible,
            "timestep": t_idx,
        }

        stop = simple_regret_best == 0
        if t_idx % cfg.log_interval == 0 or stop:
            wandb.log(metrics)
            print(f"Step {t_idx}: Loss {loss}")

        if stop:
            break
    print(f"Best solution: {best_params[0].long()}")


if __name__ == "__main__":
    main()



import torch

from holo.test_functions.elemental import (
    banded_square_matrix,
    dmp_sample_log_likelihood,
    dmp_stationary_dist,
    sample_dmp,
    sample_sparse_ergodic_transition_matrix,
)


def test_banded_square_matrix():
    ndim = 5
    bandwidth = 1
    matrix = banded_square_matrix(ndim, bandwidth)
    expected_matrix = torch.tensor(
        [
            [1, 1, 0, 0, 1],
            [1, 1, 1, 0, 0],
            [0, 1, 1, 1, 0],
            [0, 0, 1, 1, 1],
            [1, 0, 0, 1, 1],
        ]
    ).float()
    assert torch.allclose(matrix, expected_matrix)


def test_dmp_sample_log_likelihood():
    ndim = 5
    bandwidth = 1
    num_steps = 11
    num_samples = 7
    random_seed = 0
    initial_dist = torch.ones(ndim) / ndim
    generator = torch.Generator().manual_seed(random_seed)
    transition_matrix = sample_sparse_ergodic_transition_matrix(ndim, bandwidth, generator=generator)
    samples = sample_dmp(initial_dist, transition_matrix, num_steps, num_samples, generator=generator)
    log_likelihood = dmp_sample_log_likelihood(samples, initial_dist, transition_matrix)
    assert log_likelihood.shape == (num_samples,)
    assert torch.all(log_likelihood > -float("inf"))


def test_dmp_stationary_dist():
    ndim = 5
    bandwidth = 1
    random_seed = 1
    generator = torch.Generator().manual_seed(random_seed)
    transition_matrix = sample_sparse_ergodic_transition_matrix(ndim, bandwidth, generator=generator)
    stationary_dist = dmp_stationary_dist(transition_matrix)
    total_prob = stationary_dist.sum()
    assert stationary_dist.size() == (ndim, 1)
    assert torch.allclose(total_prob, torch.tensor(1.0))
    apply_transition = stationary_dist.t() @ transition_matrix
    assert torch.allclose(apply_transition, stationary_dist.t())


def test_sample_sparse_ergodic_transition_matrix():
    ndim = 5
    bandwidth = 1
    random_seed = 2
    generator = torch.Generator().manual_seed(random_seed)
    transition_matrix = sample_sparse_ergodic_transition_matrix(ndim, bandwidth, generator=generator)
    assert transition_matrix.size() == (ndim, ndim)
    assert torch.allclose(transition_matrix.sum(-1), torch.ones(ndim))


def test_sample_dmp():
    ndim = 5
    bandwidth = 1
    num_steps = 11
    num_samples = 7
    random_seed = 3
    generator = torch.Generator().manual_seed(random_seed)
    initial_dist = torch.ones(ndim) / ndim
    transition_matrix = sample_sparse_ergodic_transition_matrix(ndim, bandwidth, generator=generator)
    samples = sample_dmp(initial_dist, transition_matrix, num_steps, num_samples, generator=generator)

    assert samples.size() == (num_samples, num_steps)
    assert torch.all(samples.lt(ndim))
